@@ -22,14 +22,12 @@
  */
 
 #include "config.h"
-
 #include "empathy-roster-model-manager.h"
 
-#include "empathy-roster-model.h"
-
 #include <glib/gi18n-lib.h>
 
-#include <libempathy/empathy-utils.h>
+#include "empathy-roster-model.h"
+#include "empathy-utils.h"
 
 static void roster_model_iface_init (EmpathyRosterModelInterface *iface);
 
